from __future__ import absolute_import

from cStringIO import StringIO
from datetime import date, datetime, time
from django.core.files import File
from django.db import models
from krankshaft import valid
from tests.base import TestCaseNoDB
import base64
import tempfile
import unittest

try:
    from PIL import Image
except ImportError:
    Image = None

IMAGE_JPG = File(StringIO(base64.decodestring(''.join('''
/9j/4AAQSkZJRgABAQEAAQABAAD/2wBDAAMCAgICAgMCAgIDAwMDBAYEBAQEBAgGBgUGCQgKCgkI
CQkKDA8MCgsOCwkJDRENDg8QEBEQCgwSExIQEw8QEBD/2wBDAQMDAwQDBAgEBAgQCwkLEBAQEBAQ
EBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBD/wAARCAABAAEDASIA
AhEBAxEB/8QAFQABAQAAAAAAAAAAAAAAAAAAAAn/xAAUEAEAAAAAAAAAAAAAAAAAAAAA/8QAFAEB
AAAAAAAAAAAAAAAAAAAAAP/EABQRAQAAAAAAAAAAAAAAAAAAAAD/2gAMAwEAAhEDEQA/AKOgA//Z
'''.splitlines()))))

IMAGE_JPG_INVALID = File(StringIO(base64.decodestring(''.join('''
/9j/4AAQSkZJRgABAQEAAQABAAD/2wBDAAMCAgICAgMCAgIDAwMDBAYEBAQEBAgGBgUGCQgKCgkI
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
EBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBAQEBD/wAARCAABAAEDASIA
AhEBAxEB/8QAFQABAQAAAAAAAAAAAAAAAAAAAAn/xAAUEAEAAAAAAAAAAAAAAAAAAAAA/8QAFAEB
AAAAAAAAAAAAAAAAAAAAAP/EABQRAQAAAAAAAAAAAAAAAAAAAAD/2gAMAwEAAhEDEQA/AKOgA//Z
'''.splitlines()))))

IMAGE_PNG = File(StringIO(base64.decodestring(''.join('''
iVBORw0KGgoAAAANSUhEUgAAAAEAAAABAQMAAAAl21bKAAAABGdBTUEAALGPC/xhBQAAAAFzUkdC
AK7OHOkAAAAgY0hSTQAAeiYAAICEAAD6AAAAgOgAAHUwAADqYAAAOpgAABdwnLpRPAAAAAZQTFRF
6+vr////hyBEawAAAAFiS0dEAf8CLd4AAAAJdnBBZwAAAAIAAAACAGosfoAAAAAKSURBVAjXY2AA
AAACAAHiIbwzAAAAJXRFWHRkYXRlOmNyZWF0ZQAyMDEzLTExLTA3VDA5OjU5OjE3LTA2OjAw/JW+
VQAAACV0RVh0ZGF0ZTptb2RpZnkAMjAxMy0xMS0wN1QwOTo1OToxNy0wNjowMI3IBukAAAAASUVO
RK5CYII=
'''.splitlines()))))

IMAGE_PNG_INVALID = File(StringIO(base64.decodestring(''.join('''
iVBORw0KGgoAAAANSUhEUgAAAAEAAAABAQMAAAAl21bKAAAABGdBTUEAALGPC/xhBQAAAAFzUkdC
AK7OHOkAAAAgY0hSTQAAeiYAAICEAAD6AAAAgOgAAHUwAADqYAAAOpgAABdwnLpRPAAAAAZQTFRF
6+vr////hyBEawAAAAFiS0dEAf8CLd4AAAAJdnBBZwAAAAIAAAACAGosfoAAAAAKSURBVAjXY2AA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
VQAAACV0RVh0ZGF0ZTptb2RpZnkAMjAxMy0xMS0wN1QwOTo1OToxNy0wNjowMI3IBukAAAAASUVO
RK5CYII=
'''.splitlines()))))

class MyCharField(models.CharField):
    pass

class MyChar2Field(models.CharField):
    pass

class ValidForeign(models.Model):
    name = models.CharField(max_length=20)

class ValidForeign2(models.Model):
    name = models.CharField(max_length=20)

class ValidForeign3(models.Model):
    foreign2 = models.ForeignKey(ValidForeign2, primary_key=True)

class ValidOneToOne(models.Model):
    name = models.CharField(max_length=20)

class ValidManyToMany(models.Model):
    name = models.CharField(max_length=20)

class Valid(models.Model):
    #id = models.AutoField()
    big_integer = models.BigIntegerField()
    boolean = models.BooleanField()
    char_max_20 = models.CharField(max_length=20)
    char_max_20_choices = models.CharField(max_length=20, choices=(
        ('a', 'A'),
        ('b', 'B'),
        ('c', 'C'),
    ))
    csv_integer = models.CommaSeparatedIntegerField(max_length=20)
    date = models.DateField()
    datetime = models.DateTimeField()
    decimal = models.DecimalField(decimal_places=1, max_digits=1)
    email = models.EmailField()
    file = models.FileField(max_length=300, upload_to='file/')
    file_path = models.FilePathField(max_length=300)
    float = models.FloatField()
    generic_ip_address = models.GenericIPAddressField()
    ip_address = models.IPAddressField()
    image = models.ImageField(max_length=300, upload_to='image/')
    integer = models.IntegerField()
    integer_nullable = models.IntegerField(null=True)
    null_boolean = models.NullBooleanField()
    positive_integer = models.PositiveIntegerField()
    positive_small_integer = models.PositiveSmallIntegerField()
    slug = models.SlugField()
    small_integer = models.SmallIntegerField()
    text = models.TextField()
    time = models.TimeField()
    url = models.URLField()

    mycharfield = MyCharField(max_length=20)
    mychar2field = MyChar2Field(max_length=20)

    foreign = models.ForeignKey(ValidForeign)
    foreign_nullable = models.ForeignKey(ValidForeign2, null=True)
    foreign3 = models.ForeignKey(ValidForeign3)
    one_to_one = models.OneToOneField(ValidOneToOne)
    many_to_many = models.ManyToManyField(ValidManyToMany)
    many_to_many_blank = models.ManyToManyField(ValidManyToMany, blank=True, related_name='many_to_many_blank')

class BaseExpecterTest(TestCaseNoDB):
    def expect(self, expected, data, clean=None, **opts):
        if clean is None:
            clean = data
        assert clean == self.expecter.expect(expected, data, **opts)

    def expect_raises(self, expected, data, errors=None, **opts):
        with self.assertRaises(valid.ValueIssue) as cm:
            self.expecter.expect(expected, data, **opts)

        if errors is not None:
            assert errors == cm.exception.errors

    def setUp(self):
        self.expecter = valid.Expecter()
        self.expecter.register(MyCharField, valid.string_or_none)

class ExpecterTest(BaseExpecterTest):
    def test_expect_simple(self):
        self.expect(valid.int, 1)

    def test_expect_simple_coerce(self):
        self.expect(valid.int, '1', 1)

    def test_expect_simple_issue(self):
        self.expect_raises(valid.int, 'a', errors=["invalid literal for int() with base 10: 'a'"])

    def test_expect_dict(self):
        self.expect({'key': valid.int}, {'key': 1})

    def test_expect_dict_coerce(self):
        self.expect({'key': valid.int}, {'key': '1'}, {'key': 1})

    def test_expect_dict_data_issue(self):
        self.expect_raises({'key': valid.int}, {'key': 'a'}, errors={
            'key': ["invalid literal for int() with base 10: 'a'"]
        })

    def test_expect_dict_type_issue(self):
        self.expect_raises({}, [], errors=["Unexpected type, expected <type 'dict'>: <type 'list'>"])

    def test_expect_dict_missing_keys(self):
        self.expect_raises({'key': valid.int}, {}, errors={
            '__nonkeyerrors__': ['Missing keys: key'],
        })

    def test_expect_dict_extra_keys(self):
        self.expect_raises({'key': valid.int}, {'key': 0, 'other': 0}, errors={
            '__nonkeyerrors__': ['Extra keys: other'],
        })

    def test_expect_list_anything(self):
        self.expect([], [1])

    def test_expect_list_zero_or_more(self):
        self.expect([valid.int], [1])

    def test_expect_list_zero_or_more_zero(self):
        self.expect([valid.int], [])

    def test_expect_list_zero_or_more_more(self):
        self.expect([valid.int], [1,2,3])

    def test_expect_list_zero_or_more_coerce(self):
        self.expect([valid.int], ['1'], [1])

    def test_expect_list_exact(self):
        self.expect([valid.int, valid.int, valid.int], [1,1,1])

    def test_expect_list_zero_or_more_multiple_validation_errors(self):
        self.expect_raises([valid.int], ['a', None, 1], errors={
            0: ["invalid literal for int() with base 10: 'a'"],
            1: ['int does not accept None'],
        })

    def test_expect_list_exact_multiple_validation_errors(self):
        self.expect_raises([valid.int, valid.int, valid.int], ['a', None, 1], errors={
            0: ["invalid literal for int() with base 10: 'a'"],
            1: ['int does not accept None'],
        })

    def test_expect_list_unbalanced_lists(self):
        self.expect_raises([valid.int, valid.int, valid.int], [1], errors={
            '__nonindexerrors__': ['Expected list of length 3, saw 1']
        })

    def test_expect_options_ignore_extra_keys(self):
        self.expect({'key': valid.int}, {'key': 1, 'extra': 2}, {'key': 1}, ignore_extra_keys=True)

    def test_expect_options_ignore_missing_keys(self):
        self.expect({'key': valid.int}, {}, ignore_missing_keys=True)

    def test_expect_options_not_strict_dict(self):
        self.expect({'key': valid.int}, {'extra': 2}, {}, strict_dict=False)

    def test_expect_tuple_is_like_list(self):
        self.expect((valid.int, valid.int, valid.int), (1, 1, 1))

    def test_expect_tuple_is_like_list_zero_or_more(self):
        self.expect((valid.int,), (1, 1, 1))

    def test_expect_unhandled_type(self):
        self.assertRaises(self.expecter.ExpectedIssue, self.expecter.expect, set(), set())

class ValidatorsTest(BaseExpecterTest):
    def test_bool_false(self):
        self.expect(valid.bool, 'no', False)
        self.expect(valid.bool, '0', False)
        self.expect(valid.bool, 'false', False)
        self.expect(valid.bool, 'null', False)
        self.expect(valid.bool, 'NO', False)
        self.expect(valid.bool, '0', False)
        self.expect(valid.bool, 'FALSE', False)
        self.expect(valid.bool, 'NULL', False)

    def test_bool_true(self):
        self.expect(valid.bool, 'yes', True)
        self.expect(valid.bool, '1', True)
        self.expect(valid.bool, 'true', True)
        self.expect(valid.bool, 'YES', True)
        self.expect(valid.bool, '1', True)
        self.expect(valid.bool, 'TRUE', True)

    def test_bool_with_none(self):
        self.expect_raises(valid.bool, None,
            errors=['bool_or_none does not accept None']
        )

    def test_bool_or_none_with_none(self):
        self.expect(valid.bool_or_none, None)

    def test_choices(self):
        self.expect(valid.choice(valid.string, ('a', 'b', 'c')), 'a')

    def test_choices_not_valid_choice(self):
        self.expect_raises(valid.choice(valid.string, ('a', 'b', 'c')), 'd',
            errors=['The value is not a valid choice: d']
        )

    def test_date(self):
        self.expect(valid.date, '2013-11-06', date(2013, 11, 06))

    def test_date_with_none(self):
        self.expect_raises(valid.date, None,
            errors=['date_or_none does not accept None']
        )

    def test_date_invalid_date(self):
        self.expect_raises(valid.date, '2013-11-99',
            errors=['day is out of range for month']
        )

    def test_date_or_none_with_none(self):
        self.expect(valid.date_or_none, None)

    def test_datetime(self):
        self.expect(valid.datetime, '2013-11-06 15:51:20', datetime(2013, 11, 06, 15, 51, 20))

    def test_datetime_iso(self):
        dt = datetime.now()
        self.expect(valid.datetime, dt.isoformat(), dt)

    def test_datetime_with_none(self):
        self.expect_raises(valid.datetime, None,
            errors=['datetime_or_none does not accept None']
        )

    def test_datetime_invalid_datetime(self):
        self.expect_raises(valid.datetime, '2013-11-99 15:51:20',
            errors=['day is out of range for month']
        )

    def test_datetime_or_none_with_none(self):
        self.expect(valid.datetime_or_none, None)

    def test_django_file(self):
        self.expect(valid.django_file, File(StringIO('hello world')))

    def test_django_file_with_none(self):
        self.expect_raises(valid.django_file, None,
            errors=['django_file_or_none does not accept None']
        )

    def test_django_file_or_none_with_none(self):
        self.expect(valid.django_file_or_none, None)

    def test_django_file_invalid_file(self):
        self.expect_raises(valid.django_file, StringIO('hello world'),
            errors=['No django file found']
        )

    def test_django_file_invalid_input(self):
        self.expect_raises(valid.django_file, 'hello world',
            errors=['No django file found']
        )

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_django_image_jpg(self):
        self.expect(valid.django_image, IMAGE_JPG)

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_django_image_jpg_invalid(self):
        self.expect_raises(valid.django_image, IMAGE_JPG_INVALID,
            errors=['No valid django image (file) found']
        )

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_django_image_png(self):
        self.expect(valid.django_image, IMAGE_PNG)

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_django_image_png_invalid(self):
        self.expect_raises(valid.django_image, IMAGE_PNG_INVALID,
            errors=['No valid django image (file) found']
        )

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_django_image_with_none(self):
        self.expect_raises(valid.django_image, None,
            errors=['django_image_or_none does not accept None']
        )

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_django_image_or_none_with_none(self):
        self.expect(valid.django_image_or_none, None)

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_django_image_invalid_image(self):
        self.expect_raises(valid.django_image, StringIO('hello world'),
            errors=['No django file found']
        )

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_django_image_invalid_input(self):
        self.expect_raises(valid.django_image, 'hello world',
            errors=['No django file found']
        )

    def test_django_validator(self):
        from django.core.validators import validate_email
        validator = valid.django_validator(valid.string, validate_email)
        self.expect(validator, 'me@somewhere.com')

    def test_django_validator_invalid(self):
        from django.core.validators import validate_email
        validator = valid.django_validator(valid.string, validate_email)
        self.expect_raises(validator, 'mesomewhere.com',
            errors=['Enter a valid e-mail address.']
        )

    def test_email(self):
        self.expect(valid.email, 'me@somewhere.com')

    def test_email_with_none(self):
        self.expect_raises(valid.email, None,
            errors=['string_or_none does not accept None']
        )

    def test_email_invalid(self):
        self.expect_raises(valid.email, 'mesomewhere.com',
            errors=['Enter a valid e-mail address.']
        )

    def test_email_or_none_with_none(self):
        self.expect(valid.email_or_none, None)

    def test_float(self):
        self.expect(valid.float, '1', 1)
        self.expect(valid.float, '1.1', 1.1)
        self.expect(valid.float, '.1', .1)
        self.expect(valid.float, '0.1', .1)

    def test_float_with_none(self):
        self.expect_raises(valid.float, None, errors=['float does not accept None'])

    def test_float_invalid(self):
        self.expect_raises(valid.float, 'a', errors=['could not convert string to float: a'])
        self.expect_raises(valid.float, '1.a', errors=['invalid literal for float(): 1.a'])
        self.expect_raises(valid.float, '.a', errors=['could not convert string to float: .a'])
        self.expect_raises(valid.float, 'a.1', errors=['could not convert string to float: a.1'])

    def test_float_or_none_with_none(self):
        self.expect(valid.float_or_none, None)

    def test_int(self):
        self.expect(valid.int, 1)

    def test_int_with_invalid(self):
        self.expect_raises(valid.int, 'a', errors=["invalid literal for int() with base 10: 'a'"])

    def test_int_with_none(self):
        self.expect_raises(valid.int, None, errors=['int does not accept None'])

    def test_int_or_none(self):
        self.expect(valid.int_or_none, 1)

    def test_int_or_none_with_invalid(self):
        self.expect_raises(valid.int_or_none, 'a', errors=["invalid literal for int() with base 10: 'a'"])

    def test_int_or_none_with_none(self):
        self.expect(valid.int_or_none, None)

    def test_int_range(self):
        self.expect(valid.int_range(0, 10), 5)

    def test_int_range_unbounded_high(self):
        self.expect(valid.int_range(0, None), 5)

    def test_int_range_unbounded_high_too_low(self):
        self.expect_raises(valid.int_range(0, None), -1, errors=['The value is not within the range 0 <= -1 <= None'])

    def test_int_range_unbounded_low(self):
        self.expect(valid.int_range(None, 10), 5)

    def test_int_range_unbounded_low_too_high(self):
        self.expect_raises(valid.int_range(None, 10), 11, errors=['The value is not within the range None <= 11 <= 10'])

    def test_int_range_invalid_value(self):
        self.expect_raises(valid.int_range(0, 10), None, errors=['int does not accept None'])

    def test_int_range_invalid_range_high(self):
        self.expect_raises(valid.int_range(0, 10), 11, errors=['The value is not within the range 0 <= 11 <= 10'])

    def test_int_range_invalid_range_low(self):
        self.expect_raises(valid.int_range(0, 10), -1, errors=['The value is not within the range 0 <= -1 <= 10'])

    def test_int_range_invalid_range_coerce_high(self):
        self.expect_raises(valid.int_range(0, 10), '11', errors=['The value is not within the range 0 <= 11 <= 10'])

    def test_int_range_invalid_range_coerce_low(self):
        self.expect_raises(valid.int_range(0, 10), '-1', errors=['The value is not within the range 0 <= -1 <= 10'])

    def test_int_range_invalid_range_invalid_data(self):
        self.expect_raises(valid.int_range(0, 10), 'a', errors=["invalid literal for int() with base 10: 'a'"])

    def test_int_or_none_range(self):
        self.expect(valid.int_or_none_range(0, 10), 5)

    def test_int_or_none_range_invalid_value(self):
        self.expect(valid.int_or_none_range(0, 10), None)

    def test_int_or_none_range_invalid_range_high(self):
        self.expect_raises(valid.int_or_none_range(0, 10), 11, errors=['The value is not within the range 0 <= 11 <= 10'])

    def test_int_or_none_range_invalid_range_low(self):
        self.expect_raises(valid.int_or_none_range(0, 10), -1, errors=['The value is not within the range 0 <= -1 <= 10'])

    def test_int_or_none_range_invalid_range_coerce_high(self):
        self.expect_raises(valid.int_or_none_range(0, 10), '11', errors=['The value is not within the range 0 <= 11 <= 10'])

    def test_int_or_none_range_invalid_range_coerce_low(self):
        self.expect_raises(valid.int_or_none_range(0, 10), '-1', errors=['The value is not within the range 0 <= -1 <= 10'])

    def test_int_or_none_range_invalid_range_invalid_data(self):
        self.expect_raises(valid.int_or_none_range(0, 10), 'a', errors=["invalid literal for int() with base 10: 'a'"])

    def test_int_csv(self):
        self.expect(valid.int_csv, '1,2,3')
        self.expect(valid.int_csv, '1,2,', '1,2')
        self.expect(valid.int_csv, ',2,', '2')
        self.expect(valid.int_csv, '2')

    def test_int_csv_invalid(self):
        self.expect_raises(valid.int_csv, 'a,2,3', errors={0: ["invalid literal for int() with base 10: 'a'"]})

    def test_int_csv_with_none(self):
        self.expect_raises(valid.int_csv, None, errors=['int_no_none_csv does not accept None'])

    def test_int_csv_or_none_with_none(self):
        self.expect(valid.int_csv_or_none, None)

    def test_list_n_or_more_zero(self):
        self.expect_raises(valid.list_n_or_more(valid.int, 1), [], errors={'__nonindexerrors__': ['Expected list with 1 or more elements, saw 0']})

    def test_list_n_or_more_one(self):
        self.expect(valid.list_n_or_more(valid.int, 1), [1])

    def test_list_n_or_more_more(self):
        self.expect(valid.list_n_or_more(valid.int, 1), [1,2])

    def test_list_n_or_more_invalid_data(self):
        self.expect_raises(valid.list_n_or_more(valid.int, 1), ['a',2], errors={0: ["invalid literal for int() with base 10: 'a'"]})

    def test_list_n_or_more_invalid_n(self):
        self.assertRaises(valid.KrankshaftError, valid.list_n_or_more, valid.int, 0)

    def test_slug(self):
        self.expect(valid.slug, 'HELLO WORLD', 'hello-world')

    def test_slug_with_none(self):
        self.expect_raises(valid.slug, None, errors=['slug_or_none does not accept None'])

    def test_slug_or_none_with_none(self):
        self.expect(valid.slug_or_none, None)

    def test_string(self):
        self.expect(valid.string, 'key')

    def test_string_fail_dict(self):
        self.expect_raises(valid.string, {}, errors=["Expected string, saw: <type 'dict'>"])

    def test_string_fail_integer(self):
        self.expect_raises(valid.string, 1, errors=["Expected string, saw: <type 'int'>"])

    def test_string_fail_list(self):
        self.expect_raises(valid.string, [], errors=["Expected string, saw: <type 'list'>"])

    def test_string_with_none(self):
        self.expect_raises(valid.string, None, errors=['string_or_none does not accept None'])

    def test_string_max_length(self):
        self.expect(valid.string_max_length(1), '')

    def test_string_max_length_over_limit(self):
        self.expect_raises(valid.string_max_length(1), 'aa', errors=['The value is greater than max length 1: 2'])

    def test_string_max_length_with_none(self):
        self.expect_raises(valid.string_max_length(1), None, errors=['string_or_none does not accept None'])

    def test_string_or_none_with_none(self):
        self.expect(valid.string_or_none, None)

    def test_string_or_none_max_length(self):
        self.expect(valid.string_or_none_max_length(1), '')

    def test_string_or_none_max_length_over_limit(self):
        self.expect_raises(valid.string_or_none_max_length(1), 'aa', errors=['The value is greater than max length 1: 2'])

    def test_string_or_none_max_length_with_none(self):
        self.expect(valid.string_or_none_max_length(1), None)

    def test_time(self):
        self.expect(valid.time, '15:53:21', time(15, 53, 21))

    def test_time_with_none(self):
        self.expect_raises(valid.time, None, errors=['time_or_none does not accept None'])

    def test_time_invalid_time(self):
        self.expect_raises(valid.time, '15:53:99', errors=['second must be in 0..59'])

    def test_time_or_none_with_none(self):
        self.expect(valid.time_or_none, None)


class ValidatorsFromFieldTest(BaseExpecterTest):
    def field(self, name):
        return self.expecter.from_field(
            Valid._meta.get_field_by_name(name)[0],
        )

    def test_field_id(self):
        self.expect(self.field('id'), 1)

    def test_field_id_invalid(self):
        self.expect_raises(self.field('id'), 'a', errors=["invalid literal for int() with base 10: 'a'"])

    def test_field_id_invalid_low(self):
        self.expect_raises(self.field('id'), 0, errors=['The value is not within the range 1 <= 0 <= 2147483647'])

    def test_field_big_integer(self):
        self.expect(self.field('big_integer'), 0)

    def test_field_big_integer_invalid(self):
        self.expect_raises(self.field('big_integer'), 'a', errors=["invalid literal for int() with base 10: 'a'"])

    def test_field_big_integer_invalid_high(self):
        self.expect_raises(self.field('big_integer'), 9223372036854775808, errors=['The value is not within the range -9223372036854775808 <= 9223372036854775808 <= 9223372036854775807'])

    def test_field_big_integer_invalid_low(self):
        self.expect_raises(self.field('big_integer'), -9223372036854775809, errors=['The value is not within the range -9223372036854775808 <= -9223372036854775809 <= 9223372036854775807'])

    def test_field_boolean_0(self):
        self.expect(self.field('boolean'), 0, False)

    def test_field_boolean_1(self):
        self.expect(self.field('boolean'), 1, True)

    def test_field_boolean_no(self):
        self.expect(self.field('boolean'), 'no', False)

    def test_field_boolean_yes(self):
        self.expect(self.field('boolean'), 'yes', True)

    def test_field_boolean_with_none(self):
        self.expect_raises(self.field('boolean'), None, errors=['bool_or_none does not accept None'])

    def test_field_char_max_20_empty(self):
        self.expect(self.field('char_max_20'), '')

    def test_field_char_max_20_max(self):
        self.expect(self.field('char_max_20'), 'a' * 20)

    def test_field_char_max_20_invalid_high(self):
        self.expect_raises(self.field('char_max_20'), 'a' * 21, errors=['The value is greater than max length 20: 21'])

    def test_field_char_max_20_invalid_with_none(self):
        self.expect_raises(self.field('char_max_20'), None, errors=['string_or_none does not accept None'])

    def test_field_char_max_20_choices_empty(self):
        self.expect_raises(self.field('char_max_20_choices'), '', errors=['The value is not a valid choice: '])

    def test_field_char_max_20_choices(self):
        self.expect(self.field('char_max_20_choices'), 'a')

    def test_field_char_max_20_choices_invalid(self):
        self.expect_raises(self.field('char_max_20_choices'), 'd', errors=['The value is not a valid choice: d'])

    def test_field_char_max_20_choices_invalid_with_none(self):
        self.expect_raises(self.field('char_max_20_choices'), None, errors=['string_or_none does not accept None'])

    def test_field_csv_integer_integer(self):
        self.expect_raises(self.field('csv_integer'), 1, errors=["Expected string, saw: <type 'int'>"])

    def test_field_csv_integer_string_integer(self):
        self.expect(self.field('csv_integer'), '1')

    def test_field_csv_integer_csv_int(self):
        self.expect(self.field('csv_integer'), '1,1')

    def test_field_csv_integer_invalid(self):
        self.expect_raises(self.field('csv_integer'), '1,1,a', errors={2: ["invalid literal for int() with base 10: 'a'"]})

    def test_field_csv_integer_invalid_with_none(self):
        self.expect_raises(self.field('csv_integer'), None, errors=['int_no_none_csv does not accept None'])

    def test_field_date(self):
        self.expect(self.field('date'), '2013-10-07', date(2013, 10, 7))

    def test_field_date_invalid(self):
        self.expect_raises(self.field('date'), 'aaa', errors=['Could not parse date: aaa'])

    def test_field_date_invalid_date(self):
        self.expect_raises(self.field('date'), '2013-10-99', errors=['day is out of range for month'])

    def test_field_date_with_none(self):
        self.expect_raises(self.field('date'), None, errors=['date_or_none does not accept None'])

    def test_field_datetime(self):
        self.expect(self.field('datetime'), '2013-10-07 15:21:22', datetime(2013, 10, 7, 15, 21, 22))

    def test_field_datetime_invalid(self):
        self.expect_raises(self.field('datetime'), 'aaa', errors=['Could not parse datetime: aaa'])

    def test_field_datetime_invalid_datetime(self):
        self.expect_raises(self.field('datetime'), '2013-10-99 15:21:22', errors=['day is out of range for month'])

    def test_field_datetime_with_none(self):
        self.expect_raises(self.field('datetime'), None, errors=['datetime_or_none does not accept None'])

    def test_field_decimal(self):
        self.expect(self.field('decimal'), '1.1')

    def test_field_decimal_with_none(self):
        self.expect_raises(self.field('decimal'), None, errors=['string_or_none does not accept None'])

    def test_field_email(self):
        self.expect(self.field('email'), 'me@somewhere.com')

    def test_field_email_invalid(self):
        self.expect_raises(self.field('email'), 'mesomewhere.com', errors=['Enter a valid e-mail address.'])

    def test_field_email_with_none(self):
        self.expect_raises(self.field('email'), None, errors=['string_or_none_django_validator does not accept None'])

    def test_field_file(self):
        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write('hello world')
            tmp.seek(0)
            self.expect(self.field('file'), File(tmp))

    def test_field_file_invalid(self):
        self.expect_raises(self.field('file'), StringIO('hello world'), errors=['No django file found'])

    def test_field_file_with_none(self):
        self.expect_raises(self.field('file'), None, errors=['django_file_or_none does not accept None'])

    def test_field_file_path(self):
        self.expect(self.field('file_path'), '/a/path/')

    def test_field_float(self):
        self.expect(self.field('float'), '1.1', 1.1)

    def test_field_foreign(self):
        self.expect(self.field('foreign'), 1)

    def test_field_foreign_with_none(self):
        self.expect_raises(self.field('foreign'), None, errors=['int_or_none_range_1_to_2147483647 does not accept None'])

    def test_field_foreign_nullable(self):
        self.expect(self.field('foreign_nullable'), None)

    def test_field_foreign3(self):
        self.expect(self.field('foreign3'), 1)

    def test_field_generic_ip_address(self):
        self.expect(self.field('generic_ip_address'), '192.168.1.1')

    def test_field_many_to_many(self):
        self.expect(self.field('many_to_many'), [1, 2, 3])

    def test_field_many_to_many_at_least_one(self):
        self.expect_raises(self.field('many_to_many'), [], errors={'__nonindexerrors__': ['Expected list with 1 or more elements, saw 0']})

    def test_field_many_to_many_invalid(self):
        self.expect_raises(self.field('many_to_many'), ['a', 'b', 'c'], errors={
            0: ["invalid literal for int() with base 10: 'a'"],
            1: ["invalid literal for int() with base 10: 'b'"],
            2: ["invalid literal for int() with base 10: 'c'"]})

    def test_field_many_to_many_blank(self):
        self.expect(self.field('many_to_many_blank'), [])

    def test_field_mycharfield(self):
        self.expect(self.field('mycharfield'), 'abc')

    def test_field_mychar2field_unhandled(self):
        self.assertRaises(
            self.expecter.ExpectedIssue,
            self.expecter.from_field,
            Valid._meta.get_field_by_name('mychar2field')[0],
        )

    def test_field_ip_address(self):
        self.expect(self.field('ip_address'), '192.168.1.1')

    @unittest.skipIf(not Image, 'requires PIL/Pillow')
    def test_field_image(self):
        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write(IMAGE_PNG.read())
            IMAGE_PNG.seek(0)
            tmp.seek(0)
            self.expect(self.field('image'), File(tmp))

    def test_field_integer(self):
        self.expect(self.field('integer'), 0)

    def test_field_integer_invalid_low(self):
        self.expect_raises(self.field('integer'), -2147483649, errors=['The value is not within the range -2147483648 <= -2147483649 <= 2147483647'])

    def test_field_integer_invalid_high(self):
        self.expect_raises(self.field('integer'), 21474836471, errors=['The value is not within the range -2147483648 <= 21474836471 <= 2147483647'])

    def test_field_integer_with_none(self):
        self.expect_raises(self.field('integer'), None, errors=['int_or_none_range_-2147483648_to_2147483647 does not accept None'])

    def test_field_integer_nullable_with_none(self):
        self.expect(self.field('integer_nullable'), None)

    def test_field_null_boolean_0(self):
        self.expect(self.field('null_boolean'), 0, False)

    def test_field_null_boolean_1(self):
        self.expect(self.field('null_boolean'), 1, True)

    def test_field_null_boolean_no(self):
        self.expect(self.field('null_boolean'), 'no', False)

    def test_field_null_boolean_yes(self):
        self.expect(self.field('null_boolean'), 'yes', True)

    def test_field_null_boolean_with_none(self):
        self.expect(self.field('null_boolean'), None)

    def test_field_one_to_one(self):
        self.expect(self.field('one_to_one'), 1)

    def test_field_positive_integer_0(self):
        self.expect(self.field('positive_integer'), 0)

    def test_field_positive_integer_1(self):
        self.expect(self.field('positive_integer'), 1)

    def test_field_positive_integer_minus_1(self):
        self.expect_raises(self.field('positive_integer'), -1, errors=['The value is not within the range 0 <= -1 <= 2147483647'])

    def test_field_positive_integer_with_none(self):
        self.expect_raises(self.field('positive_integer'), None, errors=['int_or_none_range_0_to_2147483647 does not accept None'])

    def test_field_positive_integer_invalid_high(self):
        self.expect_raises(self.field('positive_integer'), 2147483648, errors=['The value is not within the range 0 <= 2147483648 <= 2147483647'])

    def test_field_positive_small_integer(self):
        self.expect(self.field('positive_small_integer'), 0)

    def test_field_positive_small_integer_minus_1(self):
        self.expect_raises(self.field('positive_small_integer'), -1, errors=['The value is not within the range 0 <= -1 <= 32767'])

    def test_field_positive_small_integer_with_none(self):
        self.expect_raises(self.field('positive_small_integer'), None, errors=['int_or_none_range_0_to_32767 does not accept None'])

    def test_field_positive_small_integer_invalid_high(self):
        self.expect_raises(self.field('positive_small_integer'), 32768, errors=['The value is not within the range 0 <= 32768 <= 32767'])

    def test_field_slug(self):
        self.expect(self.field('slug'), 'HELLO WORLD', 'hello-world')

    def test_field_small_integer(self):
        self.expect(self.field('small_integer'), 0)

    def test_field_small_integer_invalid_high(self):
        self.expect_raises(self.field('small_integer'), 32768, errors=['The value is not within the range -32768 <= 32768 <= 32767'])

    def test_field_small_integer_invalid_low(self):
        self.expect_raises(self.field('small_integer'), -32769, errors=['The value is not within the range -32768 <= -32769 <= 32767'])

    def test_field_small_integer_invalid_with_none(self):
        self.expect_raises(self.field('small_integer'), None, errors=['int_or_none_range_-32768_to_32767 does not accept None'])

    def test_field_text(self):
        self.expect(self.field('text'), 'hello world')

    def test_field_text_with_none(self):
        self.expect_raises(self.field('text'), None, errors=['string_or_none does not accept None'])

    def test_field_time(self):
        self.expect(self.field('time'), '15:38:21', time(15, 38, 21))

    def test_field_time_invalid(self):
        self.expect_raises(self.field('time'), '15:99:21', errors=['minute must be in 0..59'])

    def test_field_time_invalid_non_date(self):
        self.expect_raises(self.field('time'), 'aa', errors=['Could not parse time: aa'])

    def test_field_time_with_none(self):
        self.expect_raises(self.field('time'), None, errors=['time_or_none does not accept None'])

    def test_field_url(self):
        self.expect(self.field('url'), 'https://google.com/')



from __future__ import absolute_import

from datetime import timedelta
from django.core.cache import cache
from functools import partial
from krankshaft.auth import Auth
from krankshaft.throttle import Throttle
from tests.base import TestCaseNoDB

class FakeUser(object):
    id = 1

class ThrottleBaseTest(TestCaseNoDB):
    def setUp(self):
        request = self.make_request()
        self.throttle = Throttle(request, Auth(request))

    def test_allow(self):
        self.assertEquals(self.throttle.allow(), (True, {}))

    def test_allow_suffix(self):
        self.assertEquals(self.throttle.allow('suffix'), (True, {}))

class ThrottleRateTest(TestCaseNoDB):
    def setUp(self):
        self.request = self.make_request()
        self.auth = Auth(self.request)
        self.auth.authned = self.auth.AuthnedInterface(FakeUser())

        self.cache = cache
        self.make_throttle = partial(
            Throttle,
            bucket=timedelta(seconds=2),
            cache=self.cache,
            rate=(1, timedelta(seconds=10)),
        )
        self.throttle = self.make_throttle(self.request, self.auth)
        self.now = self.throttle.timer()
        self.throttle.timer = lambda: self.now

        # make sure cache is clear
        self.cache.clear()

    def test_allow(self, wait=None):
        wait = wait or (13, )
        self.assertEquals(self.throttle.allow(), (True, {}))

        allowed, headers = self.throttle.allow()
        self.assertEquals(allowed, False)

        nreq, nsec = self.throttle.rate
        self.assertTrue(headers['Retry-After'] in wait)

        self.throttle.timer = lambda: self.now + headers['Retry-After']
        allowed, headers = self.throttle.allow()
        self.assertEquals(allowed, True)
        self.assertTrue(not headers)

    def test_allow_anon(self):
        request = self.make_request()
        self.assertEquals(
            self.make_throttle(request, Auth(request)).allow(), (False, {})
        )

    def test_allow_default(self):
        self.throttle = Throttle(
            self.request,
            self.auth,
            cache=self.cache,
            rate=(1, timedelta(seconds=60)),
        )
        self.test_allow(wait=(67,))

    def test_allow_period_not_evenly_divisible(self):
        self.throttle = Throttle(
            self.request,
            self.auth,
            bucket=timedelta(seconds=10),
            cache=self.cache,
            rate=(1, timedelta(seconds=61)),
        )
        self.test_allow(wait=(71, 81))

    def test_allow_suffix(self, wait=13):
        # no suffix
        self.assertEquals(self.throttle.allow(), (True, {}))

        allowed, headers = self.throttle.allow()
        self.assertEquals(allowed, False)

        nreq, nsec = self.throttle.rate
        self.assertEquals(headers['Retry-After'], wait)

        self.throttle.timer = lambda: self.now + wait
        allowed, headers = self.throttle.allow()
        self.assertEquals(allowed, True)
        self.assertTrue(not headers)

        # suffix
        self.throttle.timer = lambda: self.now
        self.assertEquals(
            self.throttle.allow(suffix='suffix'), (True, {})
        )

        allowed, headers = self.throttle.allow(suffix='suffix')
        self.assertEquals(allowed, False)

        nreq, nsec = self.throttle.rate
        self.assertEquals(headers['Retry-After'], wait)

        self.throttle.timer = lambda: self.now + wait
        allowed, headers = self.throttle.allow(suffix='suffix')
        self.assertEquals(allowed, True)
        self.assertTrue(not headers)



from . import util, valid
from .exceptions import KrankshaftError, QueryIssues

class Query(object):
    '''
    Encapsulate a parsed query string to allow for application to a specific
    ORM.
    '''
    Error = KrankshaftError
    Issues = QueryIssues

    defaults = {}

    def __init__(self, qs, opts=None):
        self.opts = self.options(opts or {}, self.defaults)
        self.qs = qs

    def apply(self, query, **opts):
        raise NotImplementedError('Unable to apply query')

    def copy(self):
        '''copy() -> new Query

        Create a copy of the current query.
        '''
        return self.__class__(self.qs.copy(), self.opts.copy())

    def options(self, opts, defaults):
        '''options({}, self.defaults) -> opts

        Setup options using defaults.
        '''
        return util.valid(
            util.defaults(opts, defaults),
            self.defaults.keys()
        )

    def without(self, *names):
        '''without('order_by') -> new Query

        Filter out aspects of a query.  A way to silently remove parts of a
        query.  If you'd prefer an error, disallow certain aspects of the query
        via the options on a Query or on the apply method.
        '''
        copy = self.copy()
        for name in names:
            if name in copy.qs:
                del copy.qs[name]

        return copy

class DjangoQuery(Query):
    '''
    Allow application of a query to a Django QuerySet.

    Options:

    The following options are default False, but if set to True allow queries
    to cross relations (ie: name__other_relation_name) in that specific field.

        defer_allow_related
        filter_allow_related
        only_allow_related
        ordering_allow_related

    The following options are default None, but if set are the default value
    if not specified in the request.  You can use this to specify a default
    ordering, limit, ...

        default_defer
        default_limit
        default_only
        default_order_by

    These options control which fields may be used during filtering or ordering.
    By default, they're set to the value of Indexed (a special value on this
    class).  The Indexed value's behavior is to require at least one specified
    field in the request must be indexed (for performance reasons).  You may
    also specify the Any special value on this class which allows any of the
    fields available to be used.  Lastly, you can specify a list/tuple of the
    fields you wish to only allow to be used.

        filter
        ordering

    Filter lookups are called "Field Lookups" in Django's documentation.  They
    look like this: field__startswith=...  It's value is default to the Any
    special value allowing any lookups to be used.  However, you can specify
    a list/tuple if you wish to only allow specific lookups.

        filter_lookups

    The max_limit option defines a maximum limit that can be specified.  If the
    request asks for a larger limit, it will be an error.  If the request uses
    the limit=0 option, the max_limit will be used.  By default, the max_limit
    is None.

        max_limit

    '''

    Any = object()
    Indexed = object()

    lookups_all = (
        'exact',
        'iexact',
        'contains',
        'icontains',
        'in',
        'gt',
        'gte',
        'lt',
        'lte',
        'startswith',
        'istartswith',
        'endswith',
        'iendswith',
        'range',
        'year',
        'month',
        'day',
        'week_day',
        'isnull',
        'search',
        'regex',
        'iregex',
    )

    values_bool_true = ('1', 'yes', 'true')
    values_bool_false = ('0', 'no', 'false', 'null')
    values_bool_all = values_bool_true + values_bool_false

    defaults = {
        'defer_allow_related': False,
        'filter': Indexed,
        'filter_allow_related': False,
        'filter_lookups': Any,
        'default_defer': None,
        'default_limit': None,
        'default_only': None,
        'default_order_by': None,
        'max_limit': None,
        'max_offset': None,
        'only_allow_related': False,
        'ordering': Indexed,
        'ordering_allow_related': False,
    }

    def apply(self, query, **opts):
        '''apply(queryset) -> new queryset with query, meta

        Apply query to queryset.

        Options may be specified to change the defaults set with this class.

            query.apply(queryset, default_limit=20)

        '''
        errors = []
        opts = self.options(opts, self.opts)

        model = query.model

        filter_uses_index = False
        used_filter = False
        for name, value in self.qs.iteritems():
            if name in ('defer', 'limit', 'offset', 'only', 'order_by'):
                continue

            used_filter = True

            accessor, lookup = self.parse_field(name)
            accessor_name = '__'.join(accessor)
            field = self.get_field(model, accessor)

            if not filter_uses_index:
                filter_uses_index = self.is_indexed(field)

            if opts['filter'] is not self.Any \
               and opts['filter'] is not self.Indexed \
               and accessor_name not in opts['filter']:
                errors.append('Unallowed filter used: %s' % name)
                continue

            if not opts['filter_allow_related'] and len(accessor) > 1:
                errors.append('Cross relation filters not allowed: %s' % name)
                continue

            if lookup \
               and opts['filter_lookups'] is not self.Any \
               and lookup not in opts['filter_lookups']:
                errors.append('Unallowed filter lookup used: %s' % name)
                continue

            if lookup == 'in':
                value = value.split(',')

            elif lookup == 'range':
                value = value.split(',')
                if len(value) != 2:
                    errors.append(
                        'A range lookup requires exactly 2 params: %s' % name
                    )
                    continue

            elif lookup == 'isnull':
                try:
                    value = valid.bool(value, None)
                except ValueError as exc:
                    errors.append(
                        'An isnull lookup requires a truthy/falsey value: %s'
                        % name
                    )

            query = query.filter(**{name: value})

        if used_filter \
           and opts['filter'] is self.Indexed \
           and not filter_uses_index:
            errors.append(
                'You are required to use an indexed field in the filter'
            )


        defer = self.qs.get('defer', opts['default_defer'])
        limit = self.qs.get('limit', opts['default_limit'])
        offset = self.qs.get('offset', 0)
        only = self.qs.get('only', opts['default_only'])
        order_by = self.qs.get('order_by', opts['default_order_by'])

        if defer:
            if isinstance(defer, basestring):
                defer = defer.split(',')

            for name in defer:
                accessor, lookup = self.parse_field(name)

                if not opts['defer_allow_related'] and len(accessor) > 1:
                    errors.append(
                        'Cross relation defer not allowed: %s' % name
                    )

            query = query.defer(*defer)

        if only:
            if isinstance(only, basestring):
                only = only.split(',')

            for name in only:
                accessor, lookup = self.parse_field(name)

                if not opts['only_allow_related'] and len(accessor) > 1:
                    errors.append(
                        'Cross relation only not allowed: %s' % name
                    )

            query = query.only(*only)

        if order_by:
            if isinstance(order_by, basestring):
                order_by = order_by.split(',')

            ordering_uses_index = False
            for name in order_by:
                accessor, lookup = self.parse_field(
                    name[1:] if name.startswith('-') else name
                )
                accessor_name = '__'.join(accessor)
                field = self.get_field(model, accessor)

                if not ordering_uses_index:
                    ordering_uses_index = self.is_indexed(field)

                if opts['ordering'] is not self.Any \
                   and opts['ordering'] is not self.Indexed \
                   and accessor_name not in opts['ordering']:
                    errors.append('Unallowed ordering used: %s' % name)

                if not opts['ordering_allow_related'] and len(accessor) > 1:
                    errors.append(
                        'Cross relation ordering not allowed: %s' % name
                    )

            if opts['ordering'] is self.Indexed and not ordering_uses_index:
                errors.append(
                    'You are required to use an indexed field in the order_by'
                )

            query = query.order_by(*order_by)

        if limit:
            try:
                limit = int(limit)
                if offset:
                    offset = int(offset)
            except ValueError:
                errors.append('You must specify an integer for limit/offset')
            else:
                if offset:
                    if opts['max_offset'] and offset > opts['max_offset']:
                        errors.append(
                            'Your offset is above the maximum allowed: %s '
                            % opts['max_offset']
                        )
                        offset = 0

                    elif offset < 0:
                        errors.append('A offset must be positive')
                        offset = 0

                if opts['max_limit'] and limit > opts['max_limit']:
                    errors.append(
                        'Your limit is above the maximum allowed: %s'
                        % opts['max_limit']
                    )

                elif limit == 0:
                    limit = opts['max_limit']
                    if limit:
                        query = query[offset : offset + limit]

                elif limit < 0:
                    errors.append('A limit must be positive')

                else:
                    query = query[offset : offset + limit]

        if errors:
            raise self.Issues(*errors)

        return query, self.make_meta(limit, offset)

    def get_field(self, model, accessor):
        field = None
        first = True
        for attr in accessor:
            if first:
                field = model._meta.get_field_by_name(attr)[0]

            else:
                field = field.rel.to._meta.get_field_by_name(attr)[0]
            first = False

        return field

    def is_indexed(self, field):
        return field.db_index or field.primary_key

    def make_meta(self, limit, offset):
        meta = {
            'limit': limit,
            'next': None,
            'offset': offset,
            'previous': None,
        }

        if limit is None or offset is None:
            return meta

        base = ''
        for name, value in self.qs.iteritems():
            if name in ('limit', 'offset'):
                continue

            base += '&' if base else '?'
            base += name + '=' + value

        base += '&' if base else '?'

        meta['next'] = base + 'limit=%s&offset=%s' % (limit, offset + limit)
        if offset >= limit:
            meta['previous'] = \
                base + 'limit=%s&offset=%s' % (limit, offset - limit)

        return meta

    def parse_field(self, name):
        parsed = []
        lookup = None

        for part in name.split('__'):
            if part in self.lookups_all:
                if lookup:
                    raise self.Error(
                        'Multiple lookups found for field: %s' % name
                    )
                lookup = part

            else:
                parsed.append(part)

        return tuple(parsed), lookup

    def without(self, *names):
        copy = super(DjangoQuery, self).without(*names)

        if 'defer' in names:
            copy.opts['default_defer'] = None

        if 'limit' in names:
            copy.opts['default_limit'] = None
            copy.opts['max_limit'] = None

        if 'only' in names:
            copy.opts['default_only'] = None

        if 'order_by' in names:
            copy.opts['default_order_by'] = None

        if 'offset' in names:
            copy.opts['max_offset'] = None

        return copy



#!/usr/bin/env python

from setuptools import setup, find_packages

# see: http://bugs.python.org/issue15881
try:
    import multiprocessing
except ImportError:
    pass

install_requires = [
    'mimeparse==0.1.3',
]

try:
    tests_requires = open('requirements.txt').read().splitlines()
except IOError:
    tests_requires = []

long_description = ''
try:
    long_description = open('README.rst').read()
except IOError:
    pass

setup(
    name='krankshaft',
    version='0.2.7',
    author='Dan LaMotte',
    author_email='lamotte85@gmail.com',
    url='https://github.com/dlamotte/krankshaft',
    description='A Web API Framework (with Django, ...)',
    long_description=long_description,
    packages=find_packages('.'),
    zip_safe=False,
    install_requires=install_requires,
    license='MIT',
    tests_require=tests_requires,
    test_suite='runtests.runtests',
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities'
    ],
)



from __future__ import absolute_import

from datetime import timedelta
from django.contrib.auth.models import AnonymousUser
from django.core.cache import cache
from django.db import models
from functools import partial
from krankshaft import valid
from krankshaft.api import API as APIBase
from krankshaft.auth import Auth as AuthBase
from krankshaft.authn import Authn
from krankshaft.authz import Authz
from krankshaft.exceptions import KrankshaftError
from krankshaft.resource import DjangoModelResource
from krankshaft.throttle import Throttle
from tempfile import NamedTemporaryFile
from tests.base import TestCaseNoDB
import json
import pytest
import sys

class Auth(AuthBase):
    authn = Authn()
    authz = Authz(require_authned=False)

class API(APIBase):
    Auth = Auth


class AuthzDeny(Authz):
    def is_authorized_request(self, request, authned):
        return False

class AuthDeny(AuthBase):
    authn = Authn()
    authz = AuthzDeny()

class APIDeny(APIBase):
    Auth = AuthDeny

ThrottleOne = partial(
    Throttle,
    anon_bucket=timedelta(seconds=2),
    anon_rate=(1, timedelta(seconds=10)),
    bucket=timedelta(seconds=2),
    cache=cache,
    rate=(1, timedelta(seconds=10))
)

class FakeAPIMany(models.Model):
    char_indexed = models.CharField(max_length=20, db_index=True)

class FakeAPI(models.Model):
    char_indexed = models.CharField(max_length=20, db_index=True)
    manytomany = models.ManyToManyField(FakeAPIMany)

class FakeAPI2(models.Model):
    char_indexed = models.CharField(max_length=20, db_index=True)

class Fake1(models.Model):
    name = models.CharField(max_length=20)

class Fake2(models.Model):
    fake1 = models.ForeignKey(Fake1)

class APITest(TestCaseNoDB):
    def _pre_setup(self):
        self.api = API('v1')
        self.apid = API('v1', debug=True)
        self.api_error = API('v1', debug=True, error='custom error message')
        self.api1 = API('v1')
        self.api2 = API('v1')
        self.api2.include(self.api1)

        @self.api1
        class Fake1Resource(DjangoModelResource):
            model = Fake1

        @self.api2
        class Fake2Resource(DjangoModelResource):
            model = Fake2

        super(APITest, self)._pre_setup()

        # make sure cache is clear
        cache.clear()

    def test_abort(self):
        request = self.make_request()
        self.assertRaises(self.api.Abort, self.api.abort, request, 401)
        try:
            self.api.abort(request, 401)
        except Exception, exc:
            self.assertEquals(401, exc.response.status_code)

        response = self.api.response(request, 401)
        try:
            self.api.abort(response)
        except Exception, exc:
            self.assertEquals(response, exc.response)

        self.assertRaises(
            KrankshaftError,
            self.api.abort,
            response,
            Header=''
        )
        self.assertRaises(
            TypeError,
            self.api.abort,
            request
        )

    def test_annotate(self):
        def fakeview(request):
            return \
                hasattr(request, 'auth') \
                and isinstance(request.auth, self.api.Auth) \
                and not hasattr(request, 'user')
        fakeview = self.api(fakeview)

        request = self.make_request()
        request.user = AnonymousUser()
        self.assertEqual(hasattr(request, 'auth'), False)
        self.assertEqual(fakeview(request), True)
        self.assertEqual(hasattr(request, 'auth'), False)

    def test_auth_deny(self):
        response = self.client.get('/deny/?key=value')
        self.assertEquals(response.status_code, 401)
        self.assertTrue(not response.content)

    def test_auth_deny_decorator_only(self):
        response = self.client.get('/deny-decorator-only/?key=value')
        self.assertEquals(response.status_code, 200)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )
        self.assertEquals(
            json.loads(response.content),
            {'key': 'value'}
        )

    def test_auth_deny_manual(self):
        response = self.client.get('/deny-decorator-only-manual/')
        self.assertEquals(response.status_code, 401)
        self.assertTrue(not response.content)

    def test_deserialize_delete_get_head_options(self):
        for method in (
            self.client.delete,
            self.client.get,
            self.client.head,
            self.client.options,
        ):
            response = method('/serialize-payload/?key=value&key2=value2')
            self.assertEquals(response.status_code, 200)
            self.assertEquals(
                response['Content-Type'].split(';')[0],
                'application/json'
            )
            if method not in (self.client.head,):
                self.assertEquals(
                    json.loads(response.content),
                    {'key': 'value', 'key2': 'value2'}
                )

    def test_deserialize_post_put_form_types(self):
        for method in (
            self.client.post,
            self.client.put,
        ):
            response = method(
                '/serialize-payload/',
                'key=value&key2=value2',
                content_type='application/x-www-form-urlencoded'
            )
            self.assertEquals(response.status_code, 200)
            self.assertEquals(
                response['Content-Type'].split(';')[0],
                'application/json'
            )
            self.assertEquals(
                json.loads(response.content),
                {'key': 'value', 'key2': 'value2'}
            )

            response = method(
                '/serialize-payload/',
                {'key': 'value', 'key2': 'value2'}
            )
            self.assertEquals(response.status_code, 200)
            self.assertEquals(
                response['Content-Type'].split(';')[0],
                'application/json'
            )
            self.assertEquals(
                json.loads(response.content),
                {'key': 'value', 'key2': 'value2'}
            )

            tmp = NamedTemporaryFile()
            tmp.write('value\n')
            tmp.seek(0)
            response = method('/serialize-payload/', {'file': tmp})
            self.assertEquals(response.status_code, 200)
            self.assertEquals(
                response['Content-Type'].split(';')[0],
                'application/json'
            )
            self.assertEquals(
                json.loads(response.content),
                {'file': 'value\n'}
            )
            tmp.close()

    def test_deserialize_invalid_content(self):
        response = self.client.post(
            '/serialize-payload/',
            '!',
            content_type='application/json'
        )
        self.assertEquals(response.status_code, 400)

    def test_deserialize_invalid_content_length(self):
        request = self.make_request('POST',
            data='{"key": "value"}',
            content_type='application/json',
            CONTENT_LENGTH='a'
        )
        query, data = self.api.deserialize(request)
        self.assertTrue(not data)

    def test_deserialize_invalid_content_nonabortable(self):
        request = self.make_request('POST',
            data='!',
            content_type='application/json'
        )
        self.assertRaises(
            ValueError,
            self.api.deserialize,
            request,
            abortable=False
        )

    def test_deserialize_unsupported_content_type(self):
        response = self.client.post(
            '/serialize-payload/',
            '!',
            content_type='unsupported/content-type'
        )
        self.assertEquals(response.status_code, 406)

    def test_deserialize_unsupported_content_type_nonabortable(self):
        request = self.make_request('POST',
            data='!',
            content_type='unsupported/content-type'
        )
        self.assertRaises(
            self.api.serializer.Unsupported,
            self.api.deserialize,
            request,
            abortable=False
        )

    def test_deserialize_invalid_query_string(self):
        response = self.client.get('/serialize-payload/?key=value&invalid')
        self.assertEquals(response.status_code, 200)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )
        self.assertEquals(
            json.loads(response.content),
            {'key': 'value', 'invalid': ''}
        )

    def test_defaults_dispatch_empty(self):
        self.assertEqual(self.api.options_dispatch({}), self.api.defaults_dispatch)

    def test_defaults_dispatch_none(self):
        self.assertEqual(self.api.options_dispatch(None), self.api.defaults_dispatch)

    def test_defaults_dispatch_badopt(self):
        self.assertRaises(
            self.api.InvalidOptions,
            self.api.options_dispatch,
            {'__badopt': True}
        )

    def test_expect(self):
        self.assertEqual(
            self.api.expect({'key': valid.int}, {'key': '1'}),
            {'key': 1}
        )

    def test_expect_fail(self):
        self.assertRaises(
            self.api.ValueIssue,
            self.api.expect,
            {'key': valid.int},
            {'key': 'a'}
        )

    def test_handle_exc_abort(self):
        request = self.make_request()
        try:
            self.api.abort(request, 400)
        except Exception:
            response = self.api.handle_exc(request)

        self.assertEquals(response.status_code, 400)

    def test_handle_exc_unhandled_exception(self):
        request = self.make_request()
        try:
            {}['key']
        except Exception:
            response = self.api.handle_exc(request)

        self.assertEquals(response.status_code, 500)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )

        data = json.loads(response.content)
        self.assertEquals(data['error'], self.api.error)
        self.assertTrue('exception' not in data)
        self.assertTrue('traceback' not in data)

    def test_handle_exc_unhandled_exception_debug(self):
        request = self.make_request()
        try:
            {}['key']
        except Exception:
            response = self.apid.handle_exc(request)

        self.assertEquals(response.status_code, 500)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )

        data = json.loads(response.content)
        self.assertEquals(data['error'], self.apid.error)
        self.assertEquals(data['exception'], "KeyError: 'key'")
        self.assertTrue(data['traceback'])

    def test_handle_exc_unhandled_exception_debug_custom(self):
        request = self.make_request()
        try:
            {}['key']
        except Exception:
            response = self.apid.handle_exc(request, error='myerror')

        self.assertEquals(response.status_code, 500)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )

        data = json.loads(response.content)
        self.assertEquals(data['error'], 'myerror')
        self.assertEquals(data['exception'], "KeyError: 'key'")
        self.assertTrue(data['traceback'])

    def test_handle_exc_unhandled_exception_debug_custom_init(self):
        request = self.make_request()
        try:
            {}['key']
        except Exception:
            response = self.api_error.handle_exc(request)

        self.assertEquals(response.status_code, 500)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )

        data = json.loads(response.content)
        self.assertEquals(data['error'], self.api_error.error)
        self.assertEquals(data['exception'], "KeyError: 'key'")
        self.assertTrue(data['traceback'])

    def test_handle_exc_unhandled_exception_debug_specific(self):
        request = self.make_request()
        try:
            {}['key']
        except Exception:
            exc_info_keyerror = sys.exc_info()

        try:
            [][1]
        except Exception:
            exc_info_indexerror = sys.exc_info()
            response = self.apid.handle_exc(request, exc_info=exc_info_keyerror)

        self.assertEquals(response.status_code, 500)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )

        data = json.loads(response.content)
        self.assertEquals(data['error'], self.apid.error)
        self.assertEquals(data['exception'], "KeyError: 'key'")
        self.assertTrue(data['traceback'])

    def test_include(self):
        fake1 = Fake1.objects.create(id=1, name='fake1')
        fake2 = Fake2.objects.create(id=1, fake1=fake1)

        response = self.client.get(self.api2.reverse('fake2_single', args=(1,)))
        assert response.status_code == 200
        assert json.loads(response.content) == {
            'id': 1,
            'fake1': '/app1/api/v1/fake1/1/',
            'fake1_id': 1,
            'resource_uri': '/app2/api/v1/fake2/1/',
        }

    def test_include_api1_again(self):
        self.assertRaises(self.api2.Error, self.api2.include, self.api1)

    def test_include_self(self):
        self.assertRaises(self.api2.Error, self.api2.include, self.api2)

    def test_method_get(self):
        response = self.client.get('/only-post/')
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response['Allow'], 'POST')

    def test_method_post(self):
        response = self.client.post('/only-post/')
        self.assertEqual(response.status_code, 200)

    def test_redirect(self):
        request = self.make_request()
        for code in (301, 302):
            response = self.api.redirect(request, code, '/hello world/')
            self.assertEquals(response.status_code, code)
            self.assertEquals(response['Location'], '/hello%20world/')

    def test_redirect_abort(self):
        request = self.make_request()
        for code in (301, 302):
            try:
                self.api.abort(self.api.redirect(request, code, '/'))
            except Exception, e:
                self.assertEquals(e.response.status_code, code)
                self.assertEquals(e.response['Location'], '/')

    def test_register(self):
        self.assertRaises(self.api.Error, self.api.register, None, url=())
        self.assertRaises(self.api.Error, self.api.register, None, url=(1,2,3,4,5))
        self.assertRaises(self.api.Error, self.api.register, None, url=object())

    def test_response(self):
        response = self.api.response(
            self.make_request(),
            200,
            'content',
            Content_Type='text/plain'
        )
        self.assertEquals(response.status_code, 200)
        self.assertEquals(response.content, 'content')
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'text/plain'
        )

    def test_serialize(self):
        data = {'one': 1}
        request = self.make_request(HTTP_ACCEPT='application/json; indent=4')
        response = self.api.serialize(request, 200, data)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )
        self.assertEquals(response.content, json.dumps(data, indent=4))

    def test_serialize_default_browser_accept(self):
        data = {'one': 1}
        request = self.make_request(HTTP_ACCEPT='*/*')
        response = self.api.serialize(request, 200, data)
        assert response.status_code == 200
        assert response['Content-Type'] == 'application/json; charset=utf-8'
        assert json.loads(response.content) == {'one': 1}

    def test_serialize_force_content_type(self):
        data = {'one': 1}
        request = self.make_request(HTTP_ACCEPT='application/xml')
        response = self.api.serialize(request, 200, data,
            content_type='application/json'
        )
        self.assertEquals(response.status_code, 200)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )
        self.assertEquals(response.content, json.dumps(data))

    def test_throttle(self):
        response = self.client.get('/throttle/?key=value')
        self.assertEquals(response.status_code, 200)
        self.assertEquals(
            response['Content-Type'].split(';')[0],
            'application/json'
        )
        self.assertEquals(
            json.loads(response.content),
            {'key': 'value'}
        )

        response = self.client.get('/throttle/?key=value')
        self.assertEquals(response.status_code, 429)
        self.assertTrue(not response.content)
        self.assertEquals(response['Retry-After'], '13')

    @property
    def urls(self):
        from django.conf.urls import include, url
        return self.make_urlconf(
            url('^deny/$',
                self.api(self.view_serialize_payload, auth=AuthDeny)
            ),
            url('^deny-decorator-only/$',
                self.api(auth=AuthDeny, only=True)(self.view_serialize_payload)
            ),
            url('^deny-decorator-only-manual/$',
                self.api(only=True)(self.view_auth_manual)
            ),
            url('^only-post/$', self.api(self.view_serialize_payload, methods=('POST', ))),
            url('^serialize-payload/$', self.api(self.view_serialize_payload)),
            url('^throttle/$',
                self.api(self.view_serialize_payload, throttle=ThrottleOne)
            ),
            url('^app1/api/', include(self.api1.urls)),
            url('^app2/api/', include(self.api2.urls)),
        )

    def view_auth_manual(self, request):
        auth = self.api.auth(request, Auth=AuthDeny)
        if auth:
            return self.api.serialize(request, 200, {'authed': True})

        else:
            return auth.challenge(self.api.response(request, 401))

    def view_serialize_payload(self, request):
        query, data = self.api.deserialize(request)

        for key, value in data.items():
            if hasattr(value, 'read'):
                data[key] = value.read()

        query.update(data)

        return self.api.serialize(request, 200, query)


@pytest.mark.django_db
class APIResourceTest(TestCaseNoDB):
    def _pre_setup(self):
        self.api = API('v1')
        super(APIResourceTest, self)._pre_setup()

    def test_django_model_resource_resolve(self):
        FakeAPI.objects.create(id=1, char_indexed='value1')
        FakeAPI.objects.create(id=2, char_indexed='value2')
        FakeAPI.objects.create(id=3, char_indexed='value3')
        resource, ids = self.api.resolve([
            '/api/v1/fakeapi/2/',
            '/api/v1/fakeapi/3/',
            '/api/v1/fakeapi/1/',
        ])
        assert ids == ['2', '3', '1']

        instances = resource.fetch(*ids)
        assert [
            instance.pk
            for instance in instances
        ] == [2, 3, 1]

    def test_django_model_resource_resolve_many_resources(self):
        self.assertRaises(self.api.ResolveError, self.api.resolve, [
            '/api/v1/fakeapi/1/',
            '/api/v1/fakeapi2/1/',
        ])

    def test_django_model_resource_resolve_doesnotexist(self):
        resource, ids = self.api.resolve(['/api/v1/fakeapi/999/'])
        self.assertRaises(FakeAPI.DoesNotExist, resource.fetch, *ids)

    def test_django_model_resource_resolve_nothing(self):
        self.assertRaises(self.api.ResolveError, self.api.resolve, [])

    def test_django_model_resource_resolve_not_found(self):
        self.assertRaises(self.api.ResolveError, self.api.resolve, '/not/a/valid/path/')

    def test_django_model_resource_resolve_non_resource(self):
        self.assertRaises(self.api.ResolveError, self.api.resolve, '/resource/nomethods/')

    def test_django_model_resource_reverse(self):
        assert self.api.reverse('fakeapi_list') == '/api/v1/fakeapi/'
        assert self.api.reverse('fakeapi_single', args=(1,)) == '/api/v1/fakeapi/1/'
        assert self.api.reverse('fakeapi_set', args=('1;2;3', )) == '/api/v1/fakeapi/set/1;2;3/'

    def test_ensure_failure_to_assign_url_with_resource_with_urls(self):
        api = API('v1')

        class Resource(object):
            def router(self):
                pass

            @property
            def urls(self):
                from django.conf.urls import patterns
                return patterns('',
                    (r'^resource/other/$', api.wrap(self.router)),
                )

        self.assertRaises(api.Error, api, Resource, url=r'^resource/$')

    def test_resource_no_methods(self):
        for method in [
            getattr(self.client, name)
            for name in self.api.methods
        ]:
            response = method('/resource/nomethods/')
            self.assertEqual(response.status_code, 405)
            self.assertEqual(response['Allow'], '')

    def test_resource_get_only(self):
        response = self.client.post('/resource/get-only/')
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response['Allow'], 'GET')

    def test_resource_simple_get(self):
        response = self.client.get('/resource/simple/1/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, '1')

    def test_resource_simple_post(self):
        response = self.client.post('/resource/simple/1/')
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.content, '10')

    def test_response_with_name(self):
        from django.core.urlresolvers import reverse
        response = self.client.get(reverse(self.api.endpoint('view_name')))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'view_name')

    def test_response_with_name_with_api_reverse(self):
        response = self.client.get(self.api.reverse('view_name'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'view_name')

    def test_response_with_router(self):
        response = self.client.get('/resource/with-router/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'response')

    def test_response_with_urls(self):
        response = self.client.get('/api/v1/resource/with-urls/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'with-urls')

    def test_response_with_urls_with_router_reuse_single(self):
        response = self.client.get('/api/v1/resource/with-urls-with-router-reuse/1/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, '{"single": 1}')

    def test_response_with_urls_with_router_reuse_set(self):
        response = self.client.get('/api/v1/resource/with-urls-with-router-reuse/set/1;2;3;4;5;6/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, '{"set": [1, 2, 3, 4, 5, 6]}')

    def test_view_with_url(self):
        response = self.client.get('/api/v1/view/with-url/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'view-with-url')

    def test_view_with_url_id(self):
        response = self.client.get('/api/v1/view/with-url/1/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, 'view-with-url-1-option')

    @property
    def urls(self):
        from django.conf.urls import include, url

        api = self.api

        class ResourceNoMethods(object):
            pass

        class ResourceGetOnly(object):
            def get(self):
                pass

        class ResourceSimple(object):
            def get(self, request, id):
                return api.response(request, 200, str(id))

            def post(self, request, id):
                return api.response(request, 201, str(int(id) * 10))

        class ResourceWithRouter(object):
            def route(self, request, args, kwargs):
                return self.response(request, *args, **kwargs)

            def response(self, request, *args, **kwargs):
                return api.response(request, 200, 'response')

        @api
        class ResourceWithURLs(object):
            def router(self, request, *args, **kwargs):
                return self.api.response(request, 200, 'with-urls')

            @property
            def urls(self):
                from django.conf.urls import patterns
                return patterns('',
                    (r'^resource/with-urls/$', self.api.wrap(self.router)),
                )

        @api
        class ResourceWithURLsWithRouterReuse(object):
            def get(self, request, id):
                return self.api.serialize(request, 200, {
                    'single': int(id)
                })

            def get_set(self, request, idset):
                return self.api.serialize(request, 200, {
                    'set': [int(id) for id in idset.split(';')]
                })

            def route_set(self, request, *args, **kwargs):
                return self.api.route({
                    'get': self.get_set,
                }, request, args, kwargs)

            def route_single(self, request, *args, **kwargs):
                return self.api.route({
                    'get': self.get,
                }, request, args, kwargs)

            @property
            def urls(self):
                from django.conf.urls import patterns
                return patterns('',
                    (r'^resource/with-urls-with-router-reuse/(?P<id>\d+)/$', self.api.wrap(self.route_single)),
                    (r'^resource/with-urls-with-router-reuse/set/(?P<idset>\d[\d;]*)/$', self.api.wrap(self.route_set)),
                )

        @api(url='^view/with-url/$')
        def view_with_url(request):
            return api.response(request, 200, 'view-with-url')

        @api(url=('^view/with-url/(?P<id>\d+)/$', {'extra': 'option'}))
        def view_with_url(request, id, extra=None):
            return api.response(request, 200, 'view-with-url-%s-%s' % (id, extra))

        @api(url='^view/name/$')
        def view_name(request):
            return api.response(request, 200, 'view_name')

        @api
        class FakeAPIResource(DjangoModelResource):
            model = FakeAPI

        @api
        class FakeAPI2Resource(DjangoModelResource):
            model = FakeAPI2

        return self.make_urlconf(
            url('^resource/nomethods/$', api(ResourceNoMethods)),
            url('^resource/get-only/$', api(ResourceGetOnly)),
            url('^resource/simple/(?P<id>\d+)/$', api(ResourceSimple)),
            url('^resource/with-router/$', api(ResourceWithRouter)),
            url('^api/', include(api.urls)),
        )



from __future__ import absolute_import

from datetime import date, datetime, time, timedelta
from decimal import Decimal
from django.core.handlers.wsgi import WSGIRequest
from django.test.client import FakePayload
from krankshaft.serializer import Serializer
from tests.base import TestCaseNoDB
from urlparse import urlparse
import pytz

class SerializerConvertTest(TestCaseNoDB):
    dt = datetime(2013, 6, 8, 14, 34, 28, 121251)
    dt_expect = '2013-06-08T14:34:28.121251'
    td = timedelta(1, 2, 3)
    tz = pytz.timezone('America/Chicago')

    def do(self, obj, expect):
        self.assertEquals(expect, self.serializer.convert(obj))

    def setUp(self):
        self.serializer = Serializer()

    def test_date(self):
        self.do(self.dt.date(), self.value(self.dt_expect.split('T')[0]))

    def test_datetime(self):
        self.do(self.dt, self.value(self.dt_expect))

    def test_datetime_with_timezone(self):
        self.do(
            self.tz.normalize(self.tz.localize(self.dt)),
            self.value(self.dt_expect + '-05:00')
        )

    def test_decimal(self):
        self.do(Decimal('1.7921579150'), self.value('1.7921579150'))

    def test_dict(self):
        self.do(
            {'key': self.dt},
            {'key': self.dt_expect}
        )

    def test_list(self):
        self.do(
            [self.dt],
            [self.dt_expect]
        )

    def test_primitive(self):
        for value in self.serializer.primitive_values + (1, 1L, 1.1, 'a', u'a'):
            self.assertEquals(value, self.serializer.convert(value))

    def test_serializer_default_content_type(self):
        serializer = Serializer(
            default_content_type='application/json'
        )
        content, content_type = serializer.serialize({'key': 'value'})
        self.assertEquals(content, '{"key": "value"}')
        self.assertEquals(content_type, 'application/json')

    def test_time(self):
        self.do(self.dt.time(), self.value(self.dt_expect.split('T')[1]))

    def test_timedelta(self):
        self.do(self.td, self.value(86402.000003))

    def test_tuple(self):
        self.do(
            (self.dt,),
            [self.dt_expect]
        )

    def test_unserializable(self):
        self.assertRaises(TypeError, self.serializer.convert, object())

    def value(self, val):
        return val

class SerializerJSONTest(SerializerConvertTest):
    def do(self, obj, expect):
        self.assertEquals(expect, self.serializer.to_json(obj))

    def setUp(self):
        self.serializer = Serializer()

    def test_dict(self):
        self.do(
            {'key': self.dt},
            '{"key": "%s"}' % self.dt_expect
        )

    def test_list(self):
        self.do(
            [self.dt],
            '["%s"]' % self.dt_expect
        )

    def test_tuple(self):
        self.do(
            (self.dt,),
            '["%s"]' % self.dt_expect
        )

    def test_unserializable(self):
        self.assertRaises(TypeError, self.serializer.to_json, object())

    def value(self, val):
        if isinstance(val, basestring):
            return '"%s"' % val
        else:
            return str(val)

class SerializerSerializeTest(TestCaseNoDB):
    def setUp(self):
        self.dt = SerializerConvertTest.dt
        self.dt_expect = SerializerConvertTest.dt_expect
        self.serializer = Serializer()

        self.data = {'key': self.dt}

    def test_application_json(self):
        content, content_type = \
            self.serializer.serialize(self.data, 'application/json')
        self.assertEquals(content_type, 'application/json')
        self.assertEquals(
            content,
            '{"key": "%s"}' % self.dt_expect
        )

    def test_application_json_indent(self):
        content, content_type = \
            self.serializer.serialize(self.data, 'application/json; indent=4')
        self.assertEquals(content_type, 'application/json')
        self.assertEquals(
            content,
            '{\n    "key": "%s"\n}' % self.dt_expect
        )

    def test_application_json_indent_invalid(self):
        content, content_type = \
            self.serializer.serialize(self.data, 'application/json; indent=a')
        self.assertEquals(content_type, 'application/json')
        self.assertEquals(
            content,
            '{"key": "%s"}' % self.dt_expect
        )

    def test_application_json_q(self):
        content, content_type = \
            self.serializer.serialize(self.data, 'application/json; q=0.5')
        self.assertEquals(content_type, 'application/json')
        self.assertEquals(
            content,
            '{"key": "%s"}' % self.dt_expect
        )

    def test_default(self):
        content, content_type = \
            self.serializer.serialize(self.data)
        self.assertEquals(content_type, self.serializer.default_content_type)
        self.assertEquals(
            content,
            '{"key": "%s"}' % self.dt_expect
        )

    def test_unsupported(self):
        self.assertRaises(
            self.serializer.Unsupported,
            self.serializer.get_format,
            'unsupported/content-type'
        )

class SerializerDeserializeTest(TestCaseNoDB):
    def setUp(self):
        self.data = {'key': 'value'}
        self.serializer = Serializer()

    def test_application_json(self):
        self.assertEquals(
            self.data,
            self.serializer.deserialize('{"key": "value"}', 'application/json')
        )

    def test_application_json_indent(self):
        self.assertEquals(
            self.data,
            self.serializer \
                .deserialize('{"key": "value"}', 'application/json; indent=4')
        )

    def test_invalid_data(self):
        for content_type in self.serializer.content_types.keys():
            if content_type.startswith('multipart/'):
                continue
            self.assertRaises(
                ValueError,
                self.serializer.deserialize,
                'invalid body data',
                content_type
            )

    def test_invalid_multipart_boundary(self):
        data = 'data'
        boundary = '\xffinvalid boundary'
        parsed = urlparse('/path/')
        environ = self.client._base_environ(**{
            'CONTENT_TYPE': 'multipart/form-data; boundary=%s' % boundary,
            'CONTENT_LENGTH': len(data),
            'PATH_INFO': self.client._get_path(parsed),
            'QUERY_STRING': parsed[4],
            'REQUEST_METHOD': 'POST',
            'wsgi.input': FakePayload(data),
        })
        request = WSGIRequest(environ)

        for content_type in self.serializer.content_types.keys():
            if not content_type.startswith('multipart/'):
                continue
            self.assertRaises(
                ValueError,
                self.serializer.deserialize_request,
                request,
                request.META['CONTENT_TYPE']
            )

    def test_invalid_multipart_content_length(self):
        data = 'data'
        boundary = 'boundary'
        parsed = urlparse('/path/')
        environ = self.client._base_environ(**{
            'CONTENT_TYPE': 'multipart/form-data; boundary=%s' % boundary,
            'CONTENT_LENGTH': -1,
            'PATH_INFO': self.client._get_path(parsed),
            'QUERY_STRING': parsed[4],
            'REQUEST_METHOD': 'POST',
            'wsgi.input': FakePayload(data),
        })
        request = WSGIRequest(environ)

        for content_type in self.serializer.content_types.keys():
            if not content_type.startswith('multipart/'):
                continue
            self.assertRaises(
                ValueError,
                self.serializer.deserialize_request,
                request,
                request.META['CONTENT_TYPE']
            )

class SerializerFormatTest(TestCaseNoDB):
    def setUp(self):
        self.serializer = Serializer()

    def test_get_content_type_json(self):
        self.assertEquals(
            self.serializer.get_content_type('json'),
            'application/json'
        )



# TODO fire signal on auth failure?

from . import authn, authz

class Auth(object):
    '''
    Bind a request to this object and centralizes all Authentication and
    Authorization interfaces.

    For convenience, you may test it in a boolean way to ensure the request
    is both authenticated and authorized.

        auth = Auth(request)
        if auth:
            # request is authorized (depending on your authorization scheme
            # the request may or may not be authenticated)
            obj = ...
            if auth.is_authorized_object(obj):
                # now the request is authorized to operate on the object
            else:
                # request is not authorized to the object
        else:
            # request is not authorized (but may be authenticated)

    As an option, you can set 'authn' to a list/tuple of Authn's.  When
    initialized, it will decide which Authn to use and bind it permanently.
    '''
    AuthnedInterface = authn.AuthnedInterface

    authn = authn.AuthnDjangoBasic()
    authz = authz.AuthzReadonly()

    def __init__(self, request):
        self.authned = None
        self.request = request

        if isinstance(self.authn, (list, tuple)):
            use_authn = None
            for authn in self.authn:
                if authn.can_handle(self.request):
                    use_authn = authn
                    break
            self.authn = use_authn

    def __nonzero__(self):
        return self.is_authorized_request()

    def authenticate(self):
        '''authenticate() -> self.authned

        Process the request and test if the request is authenticated.
        '''
        if not self.authn:
            return None

        authned = self.authn.authenticate(self.request)
        if authned and self.authn.is_valid(authned):
            self.authned = self.AuthnedInterface(authned)
            return self.authned
        return None

    def challenge(self, response):
        '''challenge(response) -> response

        Update a response in flight.  Useful to add HTTP Authenticate headers.
        '''
        if self.authn:
            response = self.authn.challenge(self.request, response)
        return response

    @property
    def id(self):
        '''
        A unique identifier for a request.  If not authenticated, the
        REMOTE_ADDR of the request is used.
        '''
        if self.authned:
            return '%s-%s' % (
                self.authned.name,
                self.authned.id
            )

        else:
            return 'anon-%s' % self.request.META.get('REMOTE_ADDR', 'noaddr')

    def is_authenticated(self):
        '''is_authenticated() -> bool

        Test if the bound request is authned.

        note: depends on .authenticate() being called otherwise always False
        '''
        return bool(self.authned)

    def is_authorized_object(self, obj):
        '''is_authorized_object(obj) -> bool

        Test if authorized to operate on object.
        '''
        return self.authz.is_authorized_object(self.request, self.authned, obj)

    def is_authorized_request(self):
        '''is_authorized_request() -> bool

        Test if authorized to process request further.  No guarantees are
        placed on if .is_authorized_object() will be called ever.
        '''
        return self.authz.is_authorized_request(self.request, self.authned)

    def limit(self, query):
        '''limit(query) -> limited_query

        Pass the given query through any applicable authorization limits and
        return a new limited query that should only be visible to the
        requester.
        '''
        return self.authz.limit(self.request, self.authned, query)

    @property
    def user(self):
        if self.authned:
            return self.authned.user
        return None



# TODO caching?
# TODO stop **headers crap, make a headers object and pass that around...

from . import util
from .auth import Auth
from .exceptions import \
    Abort, KrankshaftError, InvalidOptions, ResolveError, ValueIssue
from .serializer import Serializer
from .throttle import Throttle
from .util import Annotate
from .valid import Expecter
import functools
import inspect
import logging
import sys
import traceback
import urlparse

log = logging.getLogger(__name__)

class API(object):
    '''
    Create a new API.

    apiv1 = API('v1')

    Use it as a decorator:

        @apiv1
        def view(request):
            return apiv1.serialize(request, 200, {
                'key': 'value',
            })

    Or programatically:

        @apiv1(only=True)
        def view(request):
            auth = apiv1.auth(request)
            if auth:
                return apiv1.serialize(request, 200, {
                    'authed': 'response',
                })
            else:
                return apiv1.serialize(request, 200, {
                    'un-authed': 'response',
                })
    '''

    Abort = Abort
    Auth = Auth
    Error = KrankshaftError
    Expecter = Expecter
    InvalidOptions = InvalidOptions
    ResolveError = ResolveError
    Serializer = Serializer
    Throttle = Throttle
    ValueIssue = property(lambda self: self.expecter.ValueIssue)

    defaults_dispatch = {
        'auth': True,
        'error': None,
        'methods': None,
        'only': False,
        'throttle': True,
        'throttle_suffix': None,
    }

    error = 'Internal Server Error'
    methods = (
        'get',
        'head',
        'options',
        'post',
        'put',
        'delete',
    )

    def __init__(self,
        name=''
        , debug=False
        , error=None
    ):
        '''
        Options:
            debug: enable debugging (default: False)
            error: default un-handled exception

        Example of using error:

            API(error='Error, see http://host.com/in-case-of-error/')
        '''
        self.debug = debug
        self.error = error or self.error
        self.included = []
        self.loaded = False
        self.name = name
        self.registry = []

        self.expecter = self.Expecter()
        self.serializer = self.Serializer()

    def __call__(self, view_or_resource=None, register=True, url=None, **opts):
        '''To be used as a decorator.

            @api
            def view(request):
                pass

        Or

            @api(option=value)
            def view(request):
                ...

        Or

            @api
            class MyResource(object):
                ...

        Or

            @api(option=value)
            class MyResource(object):
                ...

        For options, see wrap().  Only difference is register is default True.

        Do not use for creating resource urls.  Instead use wrap().
        '''
        return self.wrap(
            view_or_resource,
            register=register,
            url=url,
            **opts
        )

    def abort(self, response_or_request, status=None, **headers):
        '''abort(request, 400)

        Abort current execution with HTTP Response with given status.  If a
        response is given, abort current execution with given response.

            api.abort(api.response(request, 400))

        Example:

            try:
                ...
                api.abort(request, 400)
                ...
            except Exception:
                return api.handle_exc(request)

        Or use the decorator version:

            @api
            def view(request):
                ...
                api.abort(request, 400)
        '''
        if hasattr(response_or_request, 'method'):
            # request
            if status is None:
                raise TypeError(
                    'abort() requires a status when passed a response'
                )
            raise self.Abort(
                self.response(response_or_request, status, **headers)
            )

        else:
            # response
            if headers:
                raise self.Error(
                    'Cannot pass headers with given a response'
                )
            raise self.Abort(response_or_request)

    def auth(self, request, Auth=None):
        '''auth(request) -> auth

        Authenticate the current request and return an instance of Auth.
        '''
        Auth = Auth or self.Auth
        auth = Auth(request)
        auth.authenticate()
        return auth

    def challenge(self, request, auth, status=401):
        return auth.challenge(self.response(request, status))

    def deserialize(self, request, abortable=True):
        '''deserialize(request) -> query, body

        Read in the request data to native data structures.
        '''
        from django.utils.datastructures import MultiValueDict

        try:
            query = urlparse.parse_qs(
                request.META.get('QUERY_STRING', ''),
                keep_blank_values=True
            )
            query = MultiValueDict(query)

            content_type = request.META.get('CONTENT_TYPE')
            content_length = request.META.get('HTTP_CONTENT_LENGTH',
                request.META.get('CONTENT_LENGTH', 0)
            )
            try:
                content_length = int(content_length)
            except ValueError:
                content_length = 0

            if content_type and content_length > 0:
                data = self.serializer.deserialize_request(
                    request,
                    content_type
                )
            else:
                data = {}

            return (query, data)
        except ValueError:
            if abortable:
                self.abort(request, 400)
            else:
                raise

        except self.serializer.Unsupported:
            if abortable:
                self.abort(request, 406)
            else:
                raise

    def dispatch(self, view, opts, request, *args, **kwargs):
        '''dispatch(view, None, request, *args, **kwargs) -> response

        Dispatch a view function wrapping it in exception handling (support
        for api.abort()) as well as handle authenticating, throttling, ... as
        defined by opts.
        '''
        opts = self.options_dispatch(opts)

        try:
            if opts['only']:
                return view(request, *args, **kwargs)

            if opts['methods'] is not None \
               and request.method not in opts['methods'] \
               and request.method.lower() not in opts['methods']:
                return self.response(request, 405,
                    Allow=', '.join([
                        method.upper()
                        for method in opts['methods']
                    ])
                )

            auth = self.Auth if opts['auth'] is True else opts['auth']
            throttle = None
            if auth:
                auth = auth(request)
                auth.authenticate()

                if not auth:
                    return self.challenge(request, auth)

                throttle = \
                    self.Throttle \
                    if opts['throttle'] is True \
                    else opts['throttle']

                if throttle:
                    throttle = throttle(request, auth)
                    allowed, headers = \
                        throttle.allow(suffix=opts['throttle_suffix'])

                    if not allowed:
                        return self.throttled(request, **headers)

            with Annotate(request, {
                'auth': auth,
                'throttle': throttle,
            }, delete=['user']):
                return view(request, *args, **kwargs)

        except Exception:
            return self.handle_exc(request, error=opts['error'])

    def endpoint(self, name):
        '''endpoint(myview.__name__) -> 'api_v1_myview'

        Define how URLs should be named.
        '''
        n = 'api_'
        if self.name:
            n += self.name + '_'
        n += name
        return n

    def expect(self, expected, data, **opts):
        '''expect({'key': int}, {'key': '1'}) -> clean data

        In the above scenario, the returned data is:

            {'key': 1}

        Notice that the 1 goes from a string to an integer as part of the
        cleaning process.  Ideally, the returned datastructure can be fed into
        whatever it needs to be safely at this point without worrying about
        types of the values as a side-effect of just validating a proper data
        structure (no extraneous keys and proper expected values for the key).

        Simple validators as well as complex are supported.  See
        krankshaft.valid module for more details.

        Raises ValueIssue when expected does not properly validate data.
        '''
        return self.expecter.expect(expected, data, **opts)

    def extra(self, **more):
        data = {
            'api': self.name,
            'debug': self.debug,
            'stack': True,
        }
        data.update(more)
        return data

    def handle_exc(self, request, exc_info=True, error=None):
        '''handle_exc(request) -> response

        Handle arbitrary exceptions.  Serves two main purposes:

        1) needed to support abort()
        2) pass other exceptions to handler500()
        '''
        if exc_info is True:
            exc_info = sys.exc_info()

        exc, inst, tb = exc_info
        if issubclass(exc, self.Abort):
            return inst.response

        else:
            return self.handler500(request, exc_info, error=error)

    def handler500(self, request, exc_info, error=None):
        '''handler500(request, sys.exc_info())

        Returns a 500 response with error details.
        '''
        exc, inst, tb = exc_info
        error = error or self.error

        log.error(
            '%s, %s: %s',
                error,
                exc.__name__,
                inst,
            exc_info=exc_info,
            extra=self.extra(),
        )

        data = {
            'error': error,
        }

        if self.debug:
            data['exception'] = '%s: %s' % (exc.__name__, inst)
            data['traceback'] = traceback.format_exception(*exc_info)

        data = self.hook_500(data, request, exc_info)

        return self.serialize(request, 500, data)

    def hook_500(self, data, request, exc_info):
        '''hook_500(data, request, exc_info) -> data

        Convenience hook for changing data returned from a 500.
        '''
        return data

    # TODO patch vary headers...
    #   - default to "Vary: Accept", depending on authn type, do we change it to
    #     "Vary: Accept, Cookie" also?
    # TODO cache control headers...
    #   - default to "Cache-Control: no-store" and "Pragma: no-cache"?
    #   - check out tastypie.cache.SimpleCache
    def hook_response(self, response):
        '''hook_response(response) -> response

        Hook to update a response after creation.
        '''
        # TODO probably need to de-construct the Content-Type via the standard
        # and patch the header... general purpose 'http' module that implements
        # rfc standards parsing routines?
        response['Content-Type'] += '; charset=utf-8'
        return response

    def include(self, otherapi):
        '''include(otherapi)

        Include another API's views/resources in this API's lookup mechanisms.
        '''
        if otherapi is self:
            raise self.Error('Refusing to include itself')

        if otherapi in self.included:
            raise self.Error('Refusing to add api twice to included list')

        self.included.append(otherapi)

    def load(self):
        '''load()

        Resources need a bit of extra help to know when everything is considered
        loaded.  When that's the case, the resources can make mappings between
        eachother using the API registry.  This signals that the registry
        is fully loaded (typically due to the call to `api.urls`).

        You may use the load routine on your resource to do any finalization
        before they're wired into the urls.
        '''
        if self.loaded:
            return

        for view in self.registered_views:
            load = getattr(view, 'load', None)
            if load:
                load()

        self.loaded = True

    def make_resource_helper(self, klass_to_wrap, opts):
        '''make_resource_helper(klass) -> helper_instance

        This helper makes it possible to decorate a class and have it be
        connectable to Django.
        '''
        api = self

        # annotate wrapping api onto the class so it may be used without a prior
        # outside reference
        klass_to_wrap.api = api

        class Helper(object):
            __doc__ = klass_to_wrap.__doc__
            instance = klass_to_wrap()
            klass = klass_to_wrap

            def __call__(self, request, *args, **kwargs):
                view = lambda request, *args, **kwargs: \
                    api.route(self.instance, request, args, kwargs)
                return api.dispatch(view, opts, request, *args, **kwargs)

            def __getattr__(self, name):
                return getattr(self.instance, name)

        Helper.__module__ = klass_to_wrap.__module__
        Helper.__name__ = 'Helper.' + klass_to_wrap.__name__

        return Helper()

    def options_dispatch(self, opts):
        '''options_dispatch({'auth': False}) -> {'auth': False, ... defaults}

        Options:
            auth: only authed requests get through (default: True)
                optionally pass in Auth subclass
            error: message to use in case of unhandled exception
            methods: HTTP methods to allow through (default: all)
            only: only wrap the method to provide
                .abort()/unhandled-exception support
            throttle: rate-limit clients (default: True)
                optionally pass in Throttle subclass, depends on auth
            throttle_suffix: suffix the throttle key
                throttle the view seperately
        '''
        return util.valid(
            util.defaults({} if opts is None else opts, self.defaults_dispatch),
            self.defaults_dispatch.keys()
        )

    def redirect(self, request, status, location, **headers):
        '''redirect(request, 302, '/location') -> response

        Create a redirect response.
        '''
        return self.response(request, status, Location=location, **headers)

    def register(self, view, url=None):
        '''register(myview)

        Register the view with the API.

        url can be just a regex or a tuple/list of (regex, kwargs, name,
        prefix). It is essentially the same calling convention as
        django.conf.urls.url but minus the view parameter.
        '''
        if url is not None:
            if not isinstance(url, basestring) \
               and not (isinstance(url, (list, tuple)) and 1 <= len(url) <= 4):
                raise self.Error(
                    'register called with invalid url param: %r' % (url, )
                )

            if hasattr(view, 'urls'):
                raise self.Error(
                    'Will not register a resource to a url '
                    'if it has an urls attribute'
                )

        self.registry.append((view, url))

    @property
    def registered_resources(self):
        '''
        Iterate over the registered resources.

            for resource in api.registered_resources:
                ...

        Access only the resources without worrying about details of the
        registry.
        '''
        for view in self.registered_views:
            if inspect.isfunction(view):
                continue
            yield view

    @property
    def registered_views(self):
        '''
        Iterate over the registered views.

            for view in api.registered_views:
                ...

        Access the views without worrying about the registry data structure.
        '''
        for view, url in self.registry:
            yield view

        for api in self.included:
            for view, url in api.registry:
                yield view

    def resolve(self, paths):
        '''resolve([path, ...]) -> resource, [1, ...]

        Resolve a URI to an instance as given by the resource pointed to:

            from krankshaft.resource import DjangoModelResource

            @api
            class MyModel(DjangoModelResource):
                model = get_model('app', 'MyModel')

            # resource = <instance MyModelResource>
            # ids = [1]
            resource, ids = api.resolve(['/api/v1/mymodel/1/'])

        If you're using a resource that adheres to the conventions of krankshaft
        (like DjangoModelResource) then you can simply do:

            resource.fetch(*ids)

        To retrieve a list of instances for those ids.  This is required for
        linked URI representations of models to work with resources.

        The assumption is that the URL defines a single capturing expression
        around the primary key of the model.  In this case, the number 1 is
        assumably captured as either an argument or a keyword argument (name
        of the keyword does not matter).

        Order of input paths is preserved on output.  So you can do this:

            resource, ids = api.resolve(paths)
            for path, id, instance in zip(paths, ids, resource.fetch(*ids)):
                ...

        And it works as expected.
        '''
        from django.core.urlresolvers import Resolver404, resolve

        if not paths:
            raise self.ResolveError('No paths given to resolve')

        resource = None
        ids = []
        for path in paths:
            try:
                view, args, kwargs = resolve(path)
            except Resolver404:
                view = None

            view_resource = \
                getattr(view, 'resource', None) \
                or getattr(view, 'im_self', None)
            if not view or not view_resource:
                raise self.ResolveError(
                    'Unable to find a resource for path: %s' % path
                )

            ids.append(args[0] if args else kwargs.values()[0])

            if resource is None:
                resource = view_resource

            if resource != view_resource:
                raise self.ResolveError(
                    'Multiple resources found for given paths'
                )

        return resource, ids

    def response(self, request, status, content=None, **headers):
        '''response(request, 200) -> response

        Create a response object.

        Header name containing underscores will be changed to dash in order to
        make it less of a burden syntactically.

            Content_Type

        Becomes:

            Content-Type
        '''
        from django import http

        if status in (301, 302):
            location = headers.pop('Location', '')
            if status == 301:
                response = http.HttpResponsePermanentRedirect(location)
            elif status == 302:
                response = http.HttpResponseRedirect(location)

        else:
            response = http.HttpResponse(status=status)

        for name, val in headers.items():
            response[util.kw_as_header(name)] = val

        if content:
            response.content = content

        return self.hook_response(response)

    def reverse(self, name, *args, **kwargs):
        '''reverse('myendpoint') -> '/url/for/endpoint/'

        Simply a shortcut for using endpoint() on a name so it maps easily to
        the name of the endpoint you've decorated with the api.

            api = API('v1')

            @api(url='^endpoint/$')
            def endpoint(request):
                ...

            urlpatterns = patterns('',
                url('^api/', include(api.urls)),
            )

            api.reverse('endpoint') -> '/api/v1/endpoint/'

        Of course you can use the standard way by doing:

            from django.core.urlresolvers import reverse
            reverse(api.endpoint('endpoint'))

        Or hardcode it (which isn't very DRY, but...):

            reverse('api_v1_endpoint')

        '''
        from django.core.urlresolvers import reverse
        return reverse(self.endpoint(name), *args, **kwargs)

    def route(self, obj, request, args, kwargs):
        '''route(obj, request, args, kwargs) -> response

        Route a request to given obj.  If a route method exists on the object,
        simply forward control to it.  Otherwise, do a simple routing method
        based on the HTTP method of the request.

        Example:

            @api
            class SimpleResource(object):
                def all(self, request, *args, **kwargs):
                    ...

                def route(self, request, args, kwargs):
                    return self.all(request, *args, **kwargs)

        Example with default routing:

            @api
            class MethodResource(object):
                def get(self, request):
                    ...

                def post(self, request):
                    ...

        If obj is a dictionary, you can specify the handling of each method
        specifically.

            methods = {
                'post': self.post,
            }
            return api.route(methods, request, args, kwargs)

        '''
        # assume its an instance of a class
        if hasattr(obj, 'route'):
            return obj.route(request, args, kwargs)

        if isinstance(obj, dict):
            avail = obj.copy()
            for method in self.methods:
                avail.setdefault(method, None)

        else:
            avail = {
                method: getattr(obj, method, None)
                for method in self.methods
            }

        # assume its a class, route to a specific method
        method = request.method.lower()
        view = avail.get(method)

        if not view:
            return self.response(request, 405,
                Allow=', '.join([
                    method.upper()
                    for method, view in avail.iteritems()
                    if view
                ])
            )

        return view(request, *args, **kwargs)

    def serialize(self, request, status, obj,
        content_type=None
        , opts=None
        , **headers
    ):
        '''serialize(request, 200, obj) -> response

        Serialize an status and object to a response given a request.
        '''
        opts = opts or {}
        content, content_type = self.serializer.serialize(
            obj,
            content_type or request.META.get('HTTP_ACCEPT'),
            **opts
        )

        headers['Content-Type'] = content_type

        return self.response(request, status, content, **headers)

    def throttled(self, request, code=429, **headers):
        return self.response(request, code, **headers)

    def update_view(self, view):
        '''update_view(view) -> view

        Hook to make updates to a view.

        In this context, a view can be a class or a function.  In the class
        case, its considered a resource.
        '''
        # Django's way of marking a view csrf_exempt
        view.csrf_exempt = True

        return view

    @property
    def urls(self):
        '''
        Returns the list of registered endpoints.

        For example, in your urls.py:

            url('^api/', include(api.urls))

        '''
        self.load()

        urlpatterns = []
        for view, url in self.registry:
            if url:
                urlitem = (url, view, None, self.endpoint(view.__name__))
                if not isinstance(url, basestring):
                    urlitem = (url[0], view) + url[1:]

                urlpatterns.append(urlitem)

            extraurls = getattr(view, 'urls', None)
            if extraurls:
                urlpatterns.extend(extraurls)


        from django.conf.urls import include, patterns, url

        urlpatterns = patterns('', *urlpatterns)
        if self.name:
            urlpatterns = patterns('',
                url(r'^%s/' % self.name, include(urlpatterns)),
            )

        return urlpatterns

    def wrap(self, view_or_resource=None, register=False, url=None, **opts):
        '''wrap(myview) -> wrapped_view

        Wrap up a view function in an API container.

        Ideally used when setting up the urls property for resources.  ie:

            @propery
            def urls(self):
                return [
                    (r'^path/$', api.wrap(self.route_list)),
                    (r'^path/(?P<id>\d+)/$', api.wrap(self.route_object)),
                ]

        However, it has the same semantics as the decorator way to wrap a view.
        Except that this function defaults register to False (vs True for the
        api decorator).  So to wrap a view that you dont want to register:

            @api.wrap
            def myview(request):
                ...

        Options:

            register    whether or not to register the view (default: False)
            url         passed directly to register()

        See options_dispatch() for more available options.
        '''
        self.options_dispatch(opts)

        def decorator(view_or_resource):
            if inspect.isclass(view_or_resource):
                view = self.make_resource_helper(view_or_resource, opts)

            else:
                @functools.wraps(view_or_resource)
                def view(request, *args, **kwargs):
                    return self.dispatch(
                        view_or_resource,
                        opts,
                        request,
                        *args,
                        **kwargs
                    )

                # propagate these for resolve, if we've wrapped a resource
                # method, this is the only way we can find our way back to
                # the original resource (not thrilled with this hacky-ness)
                try:
                    for attr in ('im_self', 'resource'):
                        setattr(view, attr, getattr(view_or_resource, attr))
                except AttributeError:
                    pass

            view = self.update_view(view)
            if register:
                self.register(view, url=url)
            return view

        if view_or_resource:
            # used as a decorator
            # @api
            # class/def ...
            return decorator(view_or_resource)

        else:
            # passing params, return will be used as a decorator
            # @api(param=val)
            # class/def ...
            return decorator
